"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ValueCompletionProvider = exports.StateEnumCompletionProvider = exports.StateSelectorCompletionProvider = exports.StateTypeCompletionProvider = exports.PseudoElementCompletionProvider = exports.StImportNamedCompletionProvider = exports.NamedCompletionProvider = exports.FormatterCompletionProvider = exports.CodeMixinCompletionProvider = exports.CssMixinCompletionProvider = exports.ExtendCompletionProvider = exports.SelectorCompletionProvider = exports.GlobalCompletionProvider = exports.ValueDirectiveProvider = exports.TopLevelDirectiveProvider = exports.RulesetInternalDirectivesProvider = exports.ImportInternalDirectivesProvider = exports.createRange = exports.ProviderLocation = exports.ProviderRange = exports.ProviderPosition = void 0;
const path_1 = __importStar(require("path"));
const postcss_value_parser_1 = __importDefault(require("postcss-value-parser"));
const core_1 = require("@stylable/core");
const completion_types_1 = require("./completion-types");
const pseudo_class_1 = require("./feature/pseudo-class");
const provider_1 = require("./provider");
const postcss_ast_utils_1 = require("./utils/postcss-ast-utils");
const { hasOwnProperty } = Object.prototype;
class ProviderPosition {
    constructor(line, character) {
        this.line = line;
        this.character = character;
    }
}
exports.ProviderPosition = ProviderPosition;
class ProviderRange {
    constructor(start, end) {
        this.start = start;
        this.end = end;
    }
}
exports.ProviderRange = ProviderRange;
class ProviderLocation {
    constructor(uri, range) {
        this.uri = uri;
        this.range = range;
    }
}
exports.ProviderLocation = ProviderLocation;
const cssPseudoClasses = [
    'active',
    'any',
    'checked',
    'default',
    'dir()',
    'disabled',
    'empty',
    'enabled',
    'first',
    'first-child',
    'first-of-type',
    'fullscreen',
    'focus',
    'hover',
    'indeterminate',
    'in-range',
    'invalid',
    'lang()',
    'last-child',
    'last-of-type',
    'left',
    'link',
    'not()',
    'nth-child()',
    'nth-last-child()',
    'nth-last-of-type()',
    'nth-of-type()',
    'only-child',
    'only-of-type',
    'optional',
    'out-of-range',
    'read-only',
    'read-write',
    'required',
    'right',
    'root',
    'scope',
    'target',
    'valid',
    'visited',
];
// const cssPseudoElements = [
//     '::after',
//     '::before',
//     '::cue',
//     '::first-letter',
//     '::first-line',
//     '::selection',
// ]
function createRange(startLine, startPos, endline, endPos) {
    return new ProviderRange(new ProviderPosition(startLine, startPos), new ProviderPosition(endline, endPos));
}
exports.createRange = createRange;
function createDirectiveRange(position, fullLineText, lineChunkAtCursor) {
    return new ProviderRange(new ProviderPosition(position.line, Math.max(0, position.character -
        (completion_types_1.topLevelDirectives.customSelector.startsWith(fullLineText)
            ? fullLineText.length
            : lineChunkAtCursor.length))), position);
}
const importDeclarations = ['default', 'named', 'from'];
const simpleRulesetDeclarations = [
    'extends',
    'states',
    'mixin',
];
const topLevelDeclarations = [
    'root',
    'namespace',
    'vars',
    'import',
    'customSelector',
    'stScope',
    'stImport',
    'stGlobalCustomProperty',
];
// Providers
// Syntactic
// Inside :import ruleset, which is not inside media query
// If directive doesn't already exist
exports.ImportInternalDirectivesProvider = {
    provide({ parentSelector, isMediaQuery, fullLineText, position, lineChunkAtCursor, }) {
        if (parentSelector && parentSelector.selector === ':import' && !isMediaQuery) {
            const res = [];
            importDeclarations.forEach((name) => {
                if (parentSelector.nodes.every((n) => ((0, postcss_ast_utils_1.isDeclaration)(n) && completion_types_1.importDirectives[name] !== n.prop) || (0, postcss_ast_utils_1.isComment)(n)) &&
                    completion_types_1.importDirectives[name].startsWith(fullLineText.trim())) {
                    res.push((0, completion_types_1.importInternalDirective)(name, createDirectiveRange(position, fullLineText, lineChunkAtCursor)));
                }
            });
            return res;
        }
        else {
            return [];
        }
    },
};
// Inside ruleset, which is not :import or :vars
// Only inside simple selector, except -st-mixin
// If directive doesn't already exist
exports.RulesetInternalDirectivesProvider = {
    provide({ parentSelector, isMediaQuery, fullLineText, position, lineChunkAtCursor, }) {
        const res = [];
        if (parentSelector &&
            !(parentSelector.selector === ':import' || parentSelector.selector === ':vars')) {
            if (parentSelector.nodes.every((n) => ((0, postcss_ast_utils_1.isDeclaration)(n) && completion_types_1.rulesetDirectives.mixin !== n.prop) || (0, postcss_ast_utils_1.isComment)(n)) &&
                completion_types_1.rulesetDirectives.mixin.startsWith(fullLineText.trim())) {
                res.push((0, completion_types_1.rulesetInternalDirective)('mixin', createDirectiveRange(position, fullLineText, lineChunkAtCursor)));
            }
            if (this.isSimpleSelector(parentSelector.selector) && !isMediaQuery) {
                simpleRulesetDeclarations
                    .filter((d) => d !== 'mixin')
                    .forEach((name) => {
                    if (parentSelector.nodes.every((n) => ((0, postcss_ast_utils_1.isDeclaration)(n) && completion_types_1.rulesetDirectives[name] !== n.prop) ||
                        (0, postcss_ast_utils_1.isComment)(n)) &&
                        completion_types_1.rulesetDirectives[name].startsWith(fullLineText.trim())) {
                        res.push((0, completion_types_1.rulesetInternalDirective)(name, createDirectiveRange(position, fullLineText, lineChunkAtCursor)));
                    }
                });
            }
            return res;
        }
        else {
            return [];
        }
    },
    isSimpleSelector(sel) {
        return !!/^\s*\.?[\w-]*$/.test(sel); // Only a single class or element
    },
};
// Only top level
// @namespace may not repeat
exports.TopLevelDirectiveProvider = {
    provide({ parentSelector, isMediaQuery, fullLineText, position, lineChunkAtCursor, meta, astAtCursor, }) {
        if (!parentSelector) {
            if (!isMediaQuery &&
                !(astAtCursor.type === 'atrule' &&
                    astAtCursor.name === completion_types_1.topLevelDirectives.stScope.slice(1))) {
                return topLevelDeclarations
                    .filter((d) => !meta.ast.source.input.css.includes(completion_types_1.topLevelDirectives.namespace) ||
                    d !== 'namespace')
                    .filter((d) => completion_types_1.topLevelDirectives[d].startsWith(fullLineText.trim()))
                    .map((d) => (0, completion_types_1.topLevelDirective)(d, createDirectiveRange(position, fullLineText, lineChunkAtCursor)));
            }
            else {
                return [
                    (0, completion_types_1.topLevelDirective)('root', createDirectiveRange(position, fullLineText, lineChunkAtCursor)),
                ];
            }
        }
        else {
            return [];
        }
    },
};
// Inside ruleset, which is not :import
// RHS of declaration
// Declaration is not -st-directive (except -st-mixin)
// Not inside another value()
exports.ValueDirectiveProvider = {
    provide({ parentSelector, fullLineText, position }) {
        if (parentSelector &&
            !(0, provider_1.isDirective)(fullLineText) &&
            !this.isInsideValueDirective(fullLineText, position.character) &&
            fullLineText.includes(':')) {
            const parsed = (0, postcss_value_parser_1.default)(fullLineText.slice(fullLineText.indexOf(':') + 1)).nodes;
            const node = parsed[parsed.length - 1];
            if (node &&
                (node.type === 'div' ||
                    node.type === 'space' ||
                    (node.type === 'function' && !node.unclosed) ||
                    (node.type === 'word' && 'value()'.startsWith(node.value)))) {
                return [
                    (0, completion_types_1.valueDirective)(new ProviderRange(new ProviderPosition(position.line, ~fullLineText.indexOf(',')
                        ? fullLineText.lastIndexOf(',') + 1
                        : fullLineText.indexOf(':') + 1), position)),
                ];
            }
            else {
                return [];
            }
        }
        else {
            return [];
        }
    },
    isInsideValueDirective(wholeLine, pos) {
        if (!wholeLine.includes('value(')) {
            return false;
        }
        const line = wholeLine.slice(0, pos).slice(wholeLine.lastIndexOf('value('));
        let stack = 0;
        for (let i = 0; i <= line.length; i++) {
            if (line[i] === '(') {
                stack += 1;
            }
            else if (line[i] === ')') {
                stack -= 1;
            }
        }
        return stack > 0;
    },
};
// Selector level
exports.GlobalCompletionProvider = {
    provide({ parentSelector, fullLineText, position, lineChunkAtCursor, astAtCursor, }) {
        if (!parentSelector &&
            !lineChunkAtCursor.endsWith('::') &&
            !isBetweenChars(fullLineText, position, '(', ')')) {
            if (astAtCursor.type === 'root') {
                if ((0, postcss_ast_utils_1.getAtRuleByPosition)(astAtCursor, position, 'st-import')) {
                    return [];
                }
            }
            let offset = 0;
            if (fullLineText.lastIndexOf(':') !== -1) {
                if (':global()'.startsWith(lineChunkAtCursor.slice(lineChunkAtCursor.lastIndexOf(':')))) {
                    offset = lineChunkAtCursor.slice(lineChunkAtCursor.lastIndexOf(':')).length;
                }
            }
            return [
                (0, completion_types_1.globalCompletion)(new ProviderRange(new ProviderPosition(position.line, position.character - offset), position)),
            ];
        }
        else {
            return [];
        }
    },
};
// Semantic
// Selector level
// Not after :, unless entire chunk is :
exports.SelectorCompletionProvider = {
    provide({ parentSelector, fullLineText, position, lineChunkAtCursor, meta, fakes, stylable, }) {
        if (!parentSelector && (lineChunkAtCursor === ':' || !lineChunkAtCursor.endsWith(':'))) {
            const comps = [];
            comps.push(...Object.keys(meta.getAllClasses())
                .filter((c) => c !== 'root' && fakes.findIndex((f) => f.selector === '.' + c) === -1)
                .map((c) => (0, completion_types_1.classCompletion)(c, createDirectiveRange(position, fullLineText, lineChunkAtCursor))));
            comps.push(...Object.keys(meta.customSelectors).map((c) => (0, completion_types_1.classCompletion)(c, createDirectiveRange(position, fullLineText, lineChunkAtCursor), true)));
            const moreComps = meta
                .getImportStatements()
                .filter((imp) => imp.request.endsWith('st.css'))
                .reduce((acc, imp) => {
                if (acc.every((comp) => comp.label !== imp.defaultExport)) {
                    acc.push((0, completion_types_1.classCompletion)(imp.defaultExport, createDirectiveRange(position, fullLineText, lineChunkAtCursor), true));
                }
                Object.keys(imp.named).forEach((exp) => {
                    const res = stylable.resolver.resolve(meta.getSymbol(exp));
                    if (res &&
                        res._kind === 'css' &&
                        res.symbol &&
                        (res.symbol._kind === 'class' || res.symbol._kind === 'element') &&
                        acc.every((comp) => comp.label.replace('.', '') !== imp.named[exp])) {
                        acc.push((0, completion_types_1.classCompletion)(imp.named[exp], createDirectiveRange(position, fullLineText, lineChunkAtCursor)));
                    }
                });
                return acc;
            }, comps);
            return moreComps.filter((c) => c.label.startsWith(lineChunkAtCursor));
        }
        else {
            return [];
        }
    },
};
// Inside ruleset of simple selector, not :import or :vars
// RHS of -st-extends
exports.ExtendCompletionProvider = {
    provide({ lineChunkAtCursor, position, meta, stylable }) {
        if (lineChunkAtCursor.startsWith(core_1.valueMapping.extends)) {
            const value = lineChunkAtCursor.slice((core_1.valueMapping.extends + ':').length);
            const spaces = value.search(/\S|$/);
            const str = value.slice(spaces);
            const comps = [[]];
            comps.push(...Object.keys(meta.getAllClasses())
                .filter((s) => s.startsWith(str))
                .map((s) => [s, 'Local file']));
            meta.getImportStatements().forEach((i) => {
                if (i.defaultExport &&
                    i.defaultExport.startsWith(str) &&
                    i.from.endsWith('st.css')) {
                    comps.push([i.defaultExport, i.request]);
                }
            });
            meta.getImportStatements().forEach((i) => comps.push(...Object.keys(i.named)
                .filter((s) => {
                const res = stylable.resolver.resolve(meta.getSymbol(s));
                return (res &&
                    res._kind === 'css' &&
                    (res.symbol._kind === 'class' || res.symbol._kind === 'element'));
            })
                .filter((s) => s.startsWith(str))
                .map((s) => [s, i.request])));
            return comps
                .slice(1)
                .map((c) => (0, completion_types_1.extendCompletion)(c[0], c[1], new ProviderRange(new ProviderPosition(position.line, position.character - str.length), position)));
        }
        else {
            return [];
        }
    },
};
// Inside ruleset, which is not :import or :vars
// RHS of -st-extends
exports.CssMixinCompletionProvider = {
    provide({ lineChunkAtCursor, meta, position, fullLineText }) {
        if (lineChunkAtCursor.startsWith(core_1.valueMapping.mixin + ':')) {
            const { names, lastName } = (0, provider_1.getExistingNames)(fullLineText, position);
            const symbols = meta.getAllSymbols();
            return Object.keys(symbols)
                .filter((ms) => {
                const importSymbol = symbols[ms];
                return ((importSymbol._kind === 'import' &&
                    importSymbol.import.request.endsWith('.st.css')) ||
                    importSymbol._kind === 'class');
            })
                .filter((ms) => ms.startsWith(lastName))
                .filter((ms) => !names.includes(ms))
                .map((ms) => {
                const importSymbol = symbols[ms];
                return (0, completion_types_1.cssMixinCompletion)(ms, new ProviderRange(new ProviderPosition(position.line, position.character - lastName.length), position), importSymbol._kind === 'import' ? importSymbol.import.request : 'Local file');
            });
        }
        else {
            return [];
        }
    },
};
// Mixin completions
// Inside ruleset, which is not :import or :vars
// Only inside simple selector
// RHS of -st-mixin
// There is  a JS/TS import
exports.CodeMixinCompletionProvider = {
    provide({ parentSelector, meta, fullLineText, lineChunkAtCursor, position, fs, tsLangService, stylable, }) {
        if (meta
            .getImportStatements()
            .some((imp) => imp.request.endsWith('.ts') || imp.request.endsWith('.js')) &&
            !fullLineText.trim().startsWith(core_1.valueMapping.from) &&
            parentSelector &&
            lineChunkAtCursor.startsWith(core_1.valueMapping.mixin + ':')) {
            if (fullLineText.lastIndexOf('(') > fullLineText.lastIndexOf(')')) {
                return [];
            }
            const { lastName } = (0, provider_1.getExistingNames)(fullLineText, position);
            const symbols = meta.getAllSymbols();
            return Object.keys(symbols)
                .filter((ms) => symbols[ms]._kind === 'import')
                .filter((ms) => ms.startsWith(lastName))
                .filter((ms) => {
                const res = stylable.resolver.resolve(symbols[ms]);
                return res && res._kind === 'js';
            })
                .filter((ms) => isMixin(ms, meta, fs, tsLangService))
                .map((ms) => createCodeMixinCompletion(ms, lastName, position, meta));
        }
        else {
            return [];
        }
    },
};
// Inside ruleset, which is not :import
// RHS of any rule except -st-extends, -st-from
exports.FormatterCompletionProvider = {
    provide({ meta, fullLineText, parentSelector, lineChunkAtCursor, position, fs, tsLangService, stylable, }) {
        if (meta
            .getImportStatements()
            .some((imp) => imp.request.endsWith('.ts') || imp.request.endsWith('.js')) &&
            !fullLineText.trim().startsWith(core_1.valueMapping.from) &&
            !fullLineText.trim().startsWith(core_1.valueMapping.extends) &&
            !fullLineText.trim().startsWith(core_1.valueMapping.named) &&
            parentSelector &&
            ~fullLineText.indexOf(':') &&
            fullLineText.indexOf(':') < position.character &&
            !lineChunkAtCursor.startsWith(core_1.valueMapping.mixin + ':')) {
            const { lastName } = (0, provider_1.getExistingNames)(fullLineText, position);
            const symbols = meta.getAllSymbols();
            return (Object.keys(symbols)
                .filter((ms) => symbols[ms]._kind === 'import')
                .filter((ms) => ms.startsWith(lastName))
                .filter((ms) => {
                const res = stylable.resolver.resolve(symbols[ms]);
                return res && res._kind === 'js';
            })
                // .filter(ms => names.length === 0 || !~names.indexOf(ms))
                .filter((ms) => !isMixin(ms, meta, fs, tsLangService))
                .map((ms) => createCodeMixinCompletion(ms, lastName, position, meta)));
        }
        else {
            return [];
        }
    },
};
// Inside :import
// RHS of -st-named
// import exists
exports.NamedCompletionProvider = {
    provide({ parentSelector, astAtCursor, stylable, meta, position, fullLineText, src, }) {
        const { isNamedValueLine, namedValues } = (0, provider_1.getNamedValues)(src, position.line);
        if (isNamedValueLine) {
            let importName = '';
            if (parentSelector &&
                parentSelector.selector === ':import' &&
                astAtCursor.nodes &&
                astAtCursor.nodes.length) {
                importName = astAtCursor.nodes.find((n) => n.prop === core_1.valueMapping.from).value.replace(/'|"/g, '');
            }
            else if (astAtCursor.type === 'atrule' &&
                astAtCursor.name === completion_types_1.topLevelDirectives.stImport.slice(1)) {
                const parsed = (0, postcss_value_parser_1.default)(astAtCursor.params);
                parsed.walk((n) => {
                    if (n.type === 'string') {
                        importName = n.value;
                    }
                });
            }
            else {
                return [];
            }
            const comps = [[]];
            if (importName.endsWith('.st.css')) {
                const resolvedImport = this.resolveImport(importName, stylable, meta);
                if (resolvedImport) {
                    const { lastName } = (0, provider_1.getExistingNames)(fullLineText, position);
                    getNamedCSSImports(stylable, comps, resolvedImport, lastName, namedValues, meta);
                    return comps
                        .slice(1)
                        .map((c) => (0, completion_types_1.namedCompletion)(c[0], new ProviderRange(new ProviderPosition(position.line, position.character - lastName.length), new ProviderPosition(position.line, position.character)), c[1], c[2]));
                }
            }
            else if (importName.endsWith('.js')) {
                let req;
                try {
                    req = stylable.requireModule(path_1.default.join(path_1.default.dirname(meta.source), importName));
                }
                catch (e) {
                    return [];
                }
                const { lastName } = (0, provider_1.getExistingNames)(fullLineText, position);
                Object.keys(req).forEach((k) => {
                    if (typeof req[k] === 'function' && k.startsWith(lastName)) {
                        comps.push([k, importName, 'Mixin']);
                    }
                });
                return comps
                    .slice(1)
                    .map((c) => (0, completion_types_1.namedCompletion)(c[0], new ProviderRange(new ProviderPosition(position.line, position.character - lastName.length), new ProviderPosition(position.line, position.character)), c[1], c[2]));
            }
        }
        return [];
    },
    resolveImport(importName, stylable, meta) {
        return maybeResolveImport(importName, stylable, meta);
    },
};
function maybeResolveImport(importName, stylable, meta) {
    let resolvedImport = null;
    if (importName && importName.endsWith('.st.css')) {
        try {
            const imported = meta.getImportStatements().find((i) => i.request === importName);
            resolvedImport = stylable.fileProcessor.process(stylable.resolvePath(imported.context, imported.request));
        }
        catch {
            /**/
        }
    }
    return resolvedImport;
}
exports.StImportNamedCompletionProvider = {
    provide({ astAtCursor, stylable, meta, position, fullLineText, src, }) {
        const { isNamedValueLine, namedValues } = (0, provider_1.getNamedValues)(src, position.line);
        let importName = '';
        if (astAtCursor.type === 'root') {
            const atRule = (0, postcss_ast_utils_1.getAtRuleByPosition)(astAtCursor, position, 'st-import');
            const comps = [[]];
            if (isNamedValueLine && atRule) {
                const parsed = (0, postcss_value_parser_1.default)(atRule.params);
                parsed.walk((n) => {
                    if (n.type === 'string') {
                        importName = n.value;
                    }
                });
                if (importName.endsWith('.st.css')) {
                    const resolvedImport = this.resolveImport(importName, stylable, meta);
                    if (resolvedImport) {
                        const { lastName } = (0, provider_1.getExistingNames)(fullLineText, position);
                        getNamedCSSImports(stylable, comps, resolvedImport, lastName, namedValues, meta);
                        return comps
                            .slice(1)
                            .map((c) => (0, completion_types_1.namedCompletion)(c[0], new ProviderRange(new ProviderPosition(position.line, position.character - lastName.length), new ProviderPosition(position.line, position.character)), c[1], c[2]));
                    }
                }
                else if (importName.endsWith('.js')) {
                    let req;
                    try {
                        req = stylable.requireModule(path_1.default.join(path_1.default.dirname(meta.source), importName));
                    }
                    catch (e) {
                        return [];
                    }
                    const { lastName } = (0, provider_1.getExistingNames)(fullLineText, position);
                    Object.keys(req).forEach((k) => {
                        if (typeof req[k] === 'function' && k.startsWith(lastName)) {
                            comps.push([k, importName, 'Mixin']);
                        }
                    });
                    return comps
                        .slice(1)
                        .map((c) => (0, completion_types_1.namedCompletion)(c[0], new ProviderRange(new ProviderPosition(position.line, position.character - lastName.length), new ProviderPosition(position.line, position.character)), c[1], c[2]));
                }
            }
        }
        return [];
    },
    resolveImport(importName, stylable, meta) {
        return maybeResolveImport(importName, stylable, meta);
    },
};
exports.PseudoElementCompletionProvider = {
    provide({ parentSelector, resolved, resolvedElements, resolvedRoot, lastSelectoid, lineChunkAtCursor, meta, position, fullLineText, }) {
        let comps = [];
        if (!parentSelector &&
            resolved.length > 0 &&
            !isBetweenChars(fullLineText, position, '(', ')')) {
            const [firstResolvedElement] = resolvedElements;
            const secondLastNode = firstResolvedElement[firstResolvedElement.length - 2];
            let lastNode = firstResolvedElement[firstResolvedElement.length - 1] || resolvedRoot;
            if (secondLastNode &&
                lastNode.type === 'pseudo-element' &&
                core_1.nativePseudoElements.includes(lastNode.name)) {
                lastNode = secondLastNode;
            }
            const states = lastNode.resolved.reduce((acc, cur) => {
                if (cur.symbol._kind === 'class') {
                    const symbolStates = cur.symbol[core_1.valueMapping.states];
                    if (symbolStates) {
                        acc = acc.concat(Object.keys(symbolStates));
                    }
                }
                return acc;
            }, cssPseudoClasses);
            let filter = lastNode.resolved.length
                ? states.includes(lastSelectoid.replace(':', ''))
                    ? ''
                    : lastSelectoid.replace(':', '')
                : lastNode.name;
            const scope = filter
                ? (secondLastNode === null || secondLastNode === void 0 ? void 0 : secondLastNode.type) === 'pseudo-element' &&
                    core_1.nativePseudoElements.includes(secondLastNode.name)
                    ? firstResolvedElement[firstResolvedElement.length - 3]
                    : secondLastNode
                : lastNode;
            const colons = lineChunkAtCursor.match(/:*$/)[0].length;
            scope === null || scope === void 0 ? void 0 : scope.resolved.forEach((res) => {
                if (!res.symbol[core_1.valueMapping.root]) {
                    return;
                }
                comps = comps.concat(Object.keys(res.meta.getAllClasses())
                    .concat(Object.keys(res.meta.customSelectors).map((s) => s.slice(':--'.length)))
                    .filter((e) => e.startsWith(filter) && e !== 'root')
                    .map((c) => {
                    let relPath = path_1.default.relative(path_1.default.dirname(meta.source), res.meta.source);
                    if (!relPath.startsWith('.')) {
                        relPath = './' + relPath;
                    }
                    return (0, completion_types_1.pseudoElementCompletion)(c, relPath, new ProviderRange(new ProviderPosition(position.line, position.character - (filter ? filter.length + 2 : colons)), new ProviderPosition(position.line, position.character)));
                }));
            });
            let otherScope;
            const chunksSplitByPseudo = lineChunkAtCursor.split('::');
            if (!filter &&
                chunksSplitByPseudo.length > 1 &&
                chunksSplitByPseudo[chunksSplitByPseudo.length - 1] === scope.name) {
                otherScope = resolvedElements[0][resolvedElements[0].length - 2];
                filter = scope.name;
            }
            if (otherScope) {
                otherScope.resolved.forEach((res) => {
                    if (!res.symbol[core_1.valueMapping.root]) {
                        return;
                    }
                    comps = comps.concat(Object.keys(res.meta.getAllClasses())
                        .concat(Object.keys(res.meta.customSelectors).map((s) => s.slice(':--'.length)))
                        .filter((e) => e.startsWith(filter) && e !== 'root')
                        .map((c) => {
                        let relPath = path_1.default.relative(path_1.default.dirname(meta.source), res.meta.source);
                        if (!relPath.startsWith('.')) {
                            relPath = './' + relPath;
                        }
                        return (0, completion_types_1.pseudoElementCompletion)(c, relPath, new ProviderRange(new ProviderPosition(position.line, position.character -
                            (filter ? filter.length + 2 : colons)), new ProviderPosition(position.line, position.character)));
                    }));
                });
            }
        }
        return comps;
    },
};
function getNamedCSSImports(stylable, comps, resolvedImport, lastName, namedValues, meta) {
    const namedSet = new Set(namedValues);
    for (const [symbolName, symbol] of Object.entries(resolvedImport.getAllSymbols())) {
        if (symbol._kind === 'import') {
            if (symbol.name.startsWith('--')) {
                const importedVar = stylable.createResolver({}).deepResolve(symbol);
                if (importedVar &&
                    importedVar._kind === 'css' &&
                    importedVar.symbol &&
                    importedVar.symbol._kind === 'cssVar') {
                    addCompletion(symbolName, importedVar.meta, importedVar.symbol);
                }
            }
        }
        else {
            addCompletion(symbolName, meta, symbol);
        }
    }
    function addCompletion(symbolName, meta, symbol) {
        if (symbolName.slice(0, -1).startsWith(lastName) && !namedSet.has(symbolName)) {
            comps.push([
                symbolName,
                path_1.default.relative(meta.source, resolvedImport.source).slice(1).replace(/\\/g, '/'),
                createCompletionDetail(symbol),
            ]);
        }
    }
}
function createCompletionDetail(symbol) {
    switch (symbol._kind) {
        case 'class':
            return 'Stylable class';
        case 'cssVar':
            return `${symbol.global ? 'Global ' : ''}${symbol.name}`;
        case 'var':
            return symbol.text;
        case 'element':
            return 'Stylable element';
    }
}
function isNodeRule(node) {
    return node.type === 'rule';
}
function isPositionInDecl(position, decl) {
    const srcStart = decl.source && decl.source.start;
    const srcEnd = decl.source && decl.source.end;
    if (srcStart && srcEnd) {
        const srcStartLine = srcStart.line - 1;
        const srcEndLine = srcEnd.line - 1;
        const srcStartChar = srcStart.column - 1;
        const srcEndChar = srcEnd.column - 1;
        if (srcStartLine < position.line && srcEndLine > position.line) {
            return true;
        }
        else if (srcStartLine === position.line && srcEndLine === position.line) {
            if (srcStartChar <= position.character && srcEndChar >= position.character) {
                return true;
            }
        }
    }
    return false;
}
exports.StateTypeCompletionProvider = {
    provide({ astAtCursor, fullLineText, position }) {
        const acc = [];
        if (isNodeRule(astAtCursor)) {
            const declNodes = astAtCursor.nodes;
            if (declNodes) {
                const stateDeclInPos = declNodes.find((decl) => {
                    if (decl.type === 'decl' &&
                        decl.prop === core_1.valueMapping.states &&
                        isPositionInDecl(position, decl)) {
                        return true;
                    }
                    return false;
                });
                if (stateDeclInPos) {
                    const toSuggest = (0, pseudo_class_1.resolveStateTypeOrValidator)(position, fullLineText);
                    const types = Object.keys(core_1.systemValidators);
                    const input = getStateDefinitionInput(fullLineText, position);
                    // validator completion
                    if (typeof toSuggest === 'string') {
                        const validators = core_1.systemValidators[toSuggest].subValidators;
                        if (validators) {
                            const validatorNames = Object.keys(validators);
                            let relevantValidators = validatorNames.filter((t) => t.startsWith(input));
                            relevantValidators = relevantValidators.length
                                ? relevantValidators
                                : validatorNames;
                            relevantValidators.forEach((validator) => {
                                acc.push((0, completion_types_1.stateTypeCompletion)(validator, `Stylable pseudo-class ${toSuggest} validators`, new ProviderRange(new ProviderPosition(position.line, position.character - input.length), position)));
                            });
                        }
                        // type completion
                    }
                    else if (typeof toSuggest === 'boolean' && toSuggest) {
                        let relevantTypes = types.filter((t) => t.startsWith(input));
                        relevantTypes = relevantTypes.length ? relevantTypes : types;
                        relevantTypes.forEach((type) => {
                            acc.push((0, completion_types_1.stateTypeCompletion)(type, 'Stylable pseudo-class types', new ProviderRange(new ProviderPosition(position.line, position.character - input.length), position)));
                        });
                    }
                }
            }
        }
        return acc;
    },
};
exports.StateSelectorCompletionProvider = {
    provide({ parentSelector, lineChunkAtCursor, resolvedElements, resolvedRoot, target, lastSelectoid, meta, position, fullLineText, }) {
        if (!parentSelector &&
            !lineChunkAtCursor.endsWith('::') &&
            !isBetweenChars(fullLineText, position, '(', ')')) {
            let lastNode = resolvedElements[0][resolvedElements[0].length - 1] || resolvedRoot;
            if ((lastNode === null || lastNode === void 0 ? void 0 : lastNode.type) === 'pseudo-element' &&
                core_1.nativePseudoElements.includes(lastNode.name)) {
                lastNode = resolvedElements[0][resolvedElements[0].length - 2];
            }
            const chunk = Array.isArray(target.focusChunk)
                ? target.focusChunk[target.focusChunk.length - 1]
                : target.focusChunk;
            const chunkyStates = chunk && chunk.states ? chunk.states : [];
            const allStates = collectStates(lastNode);
            const newStates = lastNode.resolved.reduce((acc, cur) => {
                let relPath = path_1.default.relative(path_1.default.dirname(meta.source), cur.meta.source);
                if (!relPath.startsWith('.')) {
                    relPath = './' + relPath;
                }
                const symbol = cur.symbol;
                if (symbol._kind === 'class') {
                    const symbolStates = symbol[core_1.valueMapping.states];
                    if (symbolStates) {
                        Object.keys(symbolStates).forEach((k) => {
                            if (!acc[k] &&
                                // selectoid is a substring of current state
                                (k.slice(0, -1).startsWith(lastSelectoid.replace(':', '')) ||
                                    // selectoid is a CSS native pseudo-sclass
                                    core_1.nativePseudoClasses.includes(lastSelectoid.replace(':', '')) ||
                                    hasOwnProperty.call(allStates, lastSelectoid.replace(':', ''))) &&
                                chunkyStates.every((cs) => cs !== k)) {
                                const symbolStates = symbol[core_1.valueMapping.states];
                                const stateDef = symbolStates && symbolStates[k];
                                // if (stateDef) {
                                const stateType = stateDef && typeof stateDef === 'object' ? stateDef.type : null;
                                acc[k] = {
                                    path: meta.source === cur.meta.source ? 'Local file' : relPath,
                                    hasParam: !!stateDef,
                                    type: stateType,
                                };
                                // }
                            }
                        });
                    }
                }
                return acc;
            }, {});
            const states = Object.keys(newStates).map((k) => {
                return { name: k, state: newStates[k] };
            });
            if (states.length === 0) {
                return [];
            }
            const lastState = lastSelectoid.replace(':', '');
            const realState = hasOwnProperty.call(allStates, lastState) ||
                core_1.nativePseudoClasses.includes(lastState);
            return states.reduce((acc, st) => {
                acc.push((0, completion_types_1.stateCompletion)(st.name, st.state.path, new ProviderRange(new ProviderPosition(position.line, lastState
                    ? realState
                        ? position.character -
                            (lineChunkAtCursor.endsWith(':') ? 1 : 0)
                        : position.character -
                            (lastState.length + 1) -
                            (lineChunkAtCursor.endsWith(':') ? 1 : 0)
                    : position.character - (lineChunkAtCursor.endsWith(':') ? 1 : 0)), position), st.state.type, st.state.hasParam));
                return acc;
            }, []);
        }
        else {
            return [];
        }
    },
};
exports.StateEnumCompletionProvider = {
    provide({ meta, astAtCursor, fullLineText, lineChunkAtCursor, position, lastSelectoid, resolvedElements, resolvedRoot, }) {
        let acc = [];
        const ast = astAtCursor;
        if (!lineChunkAtCursor.endsWith('::') && (ast.type === 'root' || ast.type === 'atrule')) {
            if (lastSelectoid.startsWith(':')) {
                const stateName = lastSelectoid.slice(1);
                const lastNode = resolvedElements[0][resolvedElements[0].length - 1] || resolvedRoot;
                const resolvedStates = collectStates(lastNode);
                if (Object.keys(resolvedStates).length) {
                    const resolvedStateNode = lastNode.resolved.find((node) => {
                        const states = node.symbol[core_1.valueMapping.states];
                        return states && states[stateName] && states[stateName].type === 'enum';
                    });
                    if (resolvedStateNode) {
                        const resolvedState = resolvedStateNode.symbol[core_1.valueMapping.states][stateName];
                        let existingInput = fullLineText.slice(0, position.character);
                        existingInput = existingInput.slice(existingInput.lastIndexOf('(') + 1);
                        if (typeof resolvedState !== 'string' &&
                            (resolvedState === null || resolvedState === void 0 ? void 0 : resolvedState.arguments.every((opt) => typeof opt === 'string'))) {
                            const options = resolvedState.arguments;
                            let filteredOptions = options.filter((opt) => opt.startsWith(existingInput));
                            filteredOptions = filteredOptions.length ? filteredOptions : options;
                            let from = 'Local file';
                            if (meta.source !== resolvedStateNode.meta.source) {
                                from = path_1.default.relative(path_1.default.dirname(meta.source), resolvedStateNode.meta.source);
                                if (!from.startsWith('.')) {
                                    from = './' + from;
                                }
                            }
                            acc = filteredOptions.map((opt) => (0, completion_types_1.stateEnumCompletion)(opt, from, new ProviderRange(new ProviderPosition(position.line, position.character - existingInput.length), position)));
                        }
                    }
                }
            }
        }
        return acc;
    },
};
exports.ValueCompletionProvider = {
    provide({ fullLineText, position, meta, stylable }) {
        if ((0, provider_1.isInValue)(fullLineText, position)) {
            const inner = fullLineText
                .slice(0, fullLineText.indexOf(')', position.character) + 1)
                .slice(fullLineText
                .slice(0, fullLineText.indexOf(')', position.character) + 1)
                .lastIndexOf('('))
                .replace('(', '')
                .replace(')', '')
                .trim();
            const comps = [];
            meta.vars.forEach((v) => {
                if (v.name.startsWith(inner)) {
                    const value = (0, core_1.evalDeclarationValue)(stylable.resolver, v.text, meta, v.node);
                    comps.push((0, completion_types_1.valueCompletion)(v.name, 'Local variable', value, new ProviderRange(new ProviderPosition(position.line, position.character - inner.length), position)));
                }
            });
            const importVars = [];
            meta.getImportStatements().forEach((imp) => {
                try {
                    const resolvedPath = stylable.resolvePath((0, path_1.dirname)(meta.source), imp.request);
                    stylable.fileProcessor.process(resolvedPath).vars.forEach((v) => importVars.push({
                        name: v.name,
                        value: v.text,
                        from: imp.request,
                        node: v.node,
                    }));
                }
                catch (e) {
                    /**/
                }
            });
            importVars.forEach((v) => {
                if (v.name.startsWith(inner) &&
                    meta
                        .getImportStatements()
                        .some((imp) => Object.keys(imp.named).some((key) => key === v.name))) {
                    const value = (0, core_1.evalDeclarationValue)(stylable.resolver, v.value, meta, v.node);
                    comps.push((0, completion_types_1.valueCompletion)(v.name, v.from, value, new ProviderRange(new ProviderPosition(position.line, position.character - inner.length), position)));
                }
            });
            return comps;
        }
        else {
            return [];
        }
    },
};
function collectStates(lastNode) {
    return lastNode.resolved.reduce((acc, cur) => {
        const symbol = cur.symbol;
        if (symbol._kind === 'class') {
            const symbolStates = symbol[core_1.valueMapping.states];
            if (symbolStates) {
                Object.keys(symbolStates).forEach((k) => {
                    const symbolStates = symbol[core_1.valueMapping.states];
                    if (symbolStates && symbolStates[k] !== undefined) {
                        acc[k] = symbolStates[k];
                    }
                });
            }
        }
        return acc;
    }, {});
}
function isBetweenChars(text, position, char1, char2) {
    const posChar = position.character;
    const textToPosition = text.slice(0, posChar);
    const textfromPosition = text.slice(posChar);
    const openingParenthesisIndex = textToPosition.lastIndexOf(char1);
    const closingParenthesisIndex = textfromPosition.indexOf(char2) + posChar;
    return posChar > openingParenthesisIndex && posChar <= closingParenthesisIndex;
}
function getStateDefinitionInput(fullLineText, position) {
    let input = fullLineText.slice(0, position.character);
    const lastParenthesis = input.lastIndexOf('(');
    input = input.slice(lastParenthesis + 1);
    return input;
}
function createCodeMixinCompletion(name, lastName, position, meta) {
    const importSymbol = meta.getSymbol(name);
    if (!importSymbol || importSymbol._kind !== 'import') {
        throw new Error('expected import symbol');
    }
    return (0, completion_types_1.codeMixinCompletion)(name, new ProviderRange(new ProviderPosition(position.line, position.character - lastName.length), position), importSymbol.import.request);
}
function isMixin(name, meta, fs, tsLangService) {
    const importSymbol = meta.getSymbol(name);
    if (importSymbol._kind === 'import') {
        if (importSymbol.import.request.endsWith('.ts')) {
            const sig = (0, provider_1.extractTsSignature)(importSymbol.import.from, name, importSymbol.type === 'default', tsLangService);
            if (!sig || !sig.declaration) {
                return false;
            }
            const rtype = sig.declaration.type
                ? sig.declaration.type.getText()
                : '';
            return /(\w+.)?object/.test(rtype.trim());
        }
        if (importSymbol.import.request.endsWith('.js')) {
            return ((0, provider_1.extractJsModifierReturnType)(name, fs.readFileSync(importSymbol.import.from, 'utf8')) === 'object');
        }
    }
    return false;
}
//# sourceMappingURL=completion-providers.js.map