"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.getColorPresentation = exports.resolveDocumentColors = void 0;
const core_1 = require("@stylable/core");
const path_1 = require("path");
const vscode_languageserver_textdocument_1 = require("vscode-languageserver-textdocument");
const vscode_uri_1 = require("vscode-uri");
const completion_providers_1 = require("../completion-providers");
const provider_1 = require("../provider");
function resolveDocumentColors(stylable, cssService, document, fs) {
    const processor = stylable.fileProcessor;
    const src = document.getText();
    const filePath = vscode_uri_1.URI.parse(document.uri).fsPath;
    const res = (0, provider_1.fixAndProcess)(src, new completion_providers_1.ProviderPosition(0, 0), fs.sep === '/' ? filePath.replace(/\\/g, '/') : filePath);
    const meta = res.processed.meta;
    const colorComps = [];
    if (meta) {
        const lines = src.split('\n');
        lines.forEach((line, ind) => {
            const valueRegex = /value\((.*?)\)/g;
            let regexResult = valueRegex.exec(line);
            while (regexResult !== null) {
                const result = regexResult[1];
                const sym = meta.getSymbol(result.trim());
                let color = null;
                if (sym && sym._kind === 'var') {
                    const doc = vscode_languageserver_textdocument_1.TextDocument.create('', 'css', 0, '.gaga {border: ' +
                        (0, core_1.evalDeclarationValue)(stylable.resolver, sym.text, meta, sym.node) +
                        '}');
                    color = cssService.findColor(doc);
                }
                else if (sym && sym._kind === 'import' && sym.type === 'named') {
                    const impMeta = processor.process(stylable.resolvePath((0, path_1.dirname)(meta.source), sym.import.request));
                    const relevantVar = impMeta.vars.find((v) => v.name === sym.name);
                    if (relevantVar) {
                        const doc = vscode_languageserver_textdocument_1.TextDocument.create('', 'css', 0, '.gaga {border: ' +
                            (0, core_1.evalDeclarationValue)(stylable.resolver, 'value(' + sym.name + ')', impMeta, relevantVar.node) +
                            '}');
                        color = cssService.findColor(doc);
                    }
                }
                if (color) {
                    const range = new completion_providers_1.ProviderRange(new completion_providers_1.ProviderPosition(ind, regexResult.index +
                        regexResult[0].indexOf(regexResult[1]) -
                        'value('.length), new completion_providers_1.ProviderPosition(ind, regexResult.index +
                        regexResult[0].indexOf(regexResult[1]) +
                        result.length +
                        ')'.length));
                    colorComps.push({ color, range });
                }
                regexResult = valueRegex.exec(line);
            }
        });
        const cleanDocument = cssService.createSanitizedDocument(meta.rawAst, filePath, document.version);
        return colorComps.concat(cssService.findColors(cleanDocument));
    }
    return [];
}
exports.resolveDocumentColors = resolveDocumentColors;
function getColorPresentation(cssService, document, params) {
    const src = document.getText();
    const res = (0, provider_1.fixAndProcess)(src, new completion_providers_1.ProviderPosition(0, 0), params.textDocument.uri);
    const meta = res.processed.meta;
    const wordStart = new completion_providers_1.ProviderPosition(params.range.start.line + 1, params.range.start.character + 1);
    let noPicker = false;
    meta === null || meta === void 0 ? void 0 : meta.rawAst.walkDecls(core_1.valueMapping.named, (node) => {
        if (node &&
            ((wordStart.line === node.source.start.line &&
                wordStart.character >= node.source.start.column) ||
                wordStart.line > node.source.start.line) &&
            ((wordStart.line === node.source.end.line &&
                wordStart.character <= node.source.end.column) ||
                wordStart.line < node.source.end.line)) {
            noPicker = true;
        }
    });
    if (noPicker) {
        return [];
    }
    return cssService.getColorPresentations(document, params.color, params.range);
}
exports.getColorPresentation = getColorPresentation;
//# sourceMappingURL=color-provider.js.map